<?php

App::uses('AppComponent', 'Controller/Component');
App::uses('CakeText', 'Utility');
App::uses('BasketLib', 'EvBasket.Lib');

class BasketComponent extends AppComponent {

/**
 * the users hash
 */
	public $hash = null;

/**
 * copy of the basket
 */
	public $basketCache = null;

/**
 * Basket model - set via Component Settings
 */
	public $Basket = null;

/**
 * Basket Item model - set via Component Settings
 */
	public $BasketItem = null;

/**
 * Basket Total model - set via Component Settings
 */
	public $BasketTotal = null;

/**
 * Basket Data model - set via Component Settings
 */
	public $BasketData = null;

/**
 * setup the cookie component
 *
 */
	public function initialize(Controller $controller) {
		parent::initialize($controller);

		if (empty($this->_controller->BasketCookie)) {
			$this->_controller->BasketCookie = $this->_controller->loadComponent('EvBasket.BasketCookie');
		}
	}

/**
 * setup a hash or retrieve a hash from the cookie
 *
 * @return string  	Basket hash identifier
 */
	public function setupHash() {
		if (! empty($this->hash)) {
			return $this->hash;
		}

		$cookieHash = $this->_controller->BasketCookie->read('basketHash');

		$userHash = $this->getBasketHashFromUser();

		// we have a cookie hash and a separate user hash that don't match
		if (! empty($cookieHash)) {
			$authUserId = $this->_controller->Auth->user('User.id');

			if (! empty($authUserId) && $userHash === false) {
				// associate the current cookie hash with the logged in user
				$this->Basket->assignUserToBasket($cookieHash, $authUserId);
				$this->hash = $cookieHash;
			} elseif ($userHash !== false && $userHash !== $cookieHash) {
				// there is a user basket hash that doesn't match the cookie one so we need to merge
				// the baskets, deletes the basket entry not associated with the user
				$this->hash = $this->Basket->mergeBaskets($userHash, $cookieHash, true);
			} else {
				$this->hash = $cookieHash;
			}
		} elseif ($userHash !== false) {
			// no cookie hash is assigned but we do have a hash for the user
			$this->hash = $userHash;
		} else {
			// We don't have any sort of basket so generate a new one
			$this->hash = $this->generateHash();
			$this->setupBasket($this->hash);
		}

		$this->_controller->BasketCookie->write(
			'basketHash',
			$this->hash
		);

		return $this->hash;
	}

/**
 * Checks the logged in user to see if they have a basket hash set.
 * @return  mixed If there is no user logged in it will return false. If there is a hash stored against the user then the hash is returned.
 */
	public function getBasketHashFromUser() {
		// If config is not set to associate users with baskets always return false without checking
		if (Configure::read('EvBasket.associateBasketWithUser') == false) {
			return false;
		}

		$userId = $this->_controller->Auth->user('User.id');

		// No user so no user basket
		if (!$userId) {
			return false;
		}

		$basket = $this->Basket->find('first', array(
			'fields' => 'hash',
			'conditions' => array(
				'user_id' => $userId
			),
			'order' => array(
				'created DESC'
			)
		));

		if ($basket) {
			return $basket['Basket']['hash'];
		}

		return false;
	}

/**
 * given the hash, setup or return the existing basket
 *
 * @param 	string 	$hash 	Basket Hash
 * @return 	bool 	$result True if a basket was created, false if one already exists
 */
	public function setupBasket($hash) {
		if (! $this->Basket->checkExists($hash)) {
			// Basket doesn't already exist so lets create one
			if (Configure::read('EvBasket.associateBasketWithUser') && $this->_controller->Auth->user()) {
				$userId = $this->_controller->Auth->user('User.id');
			} else {
				$userId = null;
			}

			return $this->Basket->createBasket(
				$hash,
				CakeSession::read('EvCurrency.currencyId'),
				$userId
			);
		}

		return false;
	}

/**
 * build the basket item contain list
 *
 * @param 	array  	$itemContain 	Value from Configure::read('EvBasket.BasketItemContains')
 * @param 	array  	$relationships 	Value from Configure::read('EvBasket.BasketItemBelongsTo')
 * @return 	array
 */
	public function buildItemContain($itemContain, $relationships) {
		if (empty($itemContain)) {
			$itemContain = array();

			foreach ($relationships as $key => $value) {
				if (is_array($value)) {
					$itemContain[] = $key;
				} else {
					$itemContain[] = $value;
				}
			}
		}

		return $itemContain;
	}

/**
 * @codeCoverageIgnore
 * generate a hash for the basket
 * extracted into another method for easier unit testing
 *
 * @return string
 */
	public function generateHash() {
		return CakeText::uuid();
	}

/**
 * build the basket item key
 *
 * @param 	array 			$item 	Array with two elements of model and model_id
 * @return 	string|bool 	$key 	Item basket key Model.Model_id format / false on fail
 */
	public function buildItemKey($item) {
		if (empty($item['model']) || empty($item['model_id'])) {
			return false;
		}

		return $item['model'] . '.' . $item['model_id'];
	}

/**
 * find item key
 *
 * @param 	string|int|array 	$key 	Either Basket row ID / compiled basket key string / Array with two elements of model and model_id
 * @return  string
 */
	public function getItemKey($key) {
		// set the key as the itemKey as a fallback
		$itemKey = $key;

		if (is_array($key)) {
			// It's an array, try and build it
			$itemKey = $this->buildItemKey($key);
		} elseif (is_numeric($key) && ! empty($this->basketCache)) {
			// it's and id number - try to find the itemKey
			$keysArray = Hash::combine(
				$this->basketCache,
				'BasketItem.{s}.id',
				array(
					'%s.%s',
					'BasketItem.{s}.model',
					'BasketItem.{s}.model_id'
				)
			);

			if (empty($keysArray[$key])) {
				return false;
			}

			$itemKey = $keysArray[$key];
		}

		return $itemKey;
	}

/**
 * given the item key, check to see if it exists already in the basket
 *
 * @param 	string 		$key 		Item basket key Model.Model_id format
 * @param 	array|null 	$basket 	Optional basket to search in - if null will use $this->basket
 * @return 	bool
 */
	public function basketItemExists($key, $basket = null) {
		if (! empty($basket)) {
			return ! empty($basket['BasketItem'][$key]);
		}

		return ! empty($this->basketCache['BasketItem'][$key]);
	}

/**
 * common method to call out to rebuild the totals
 *
 * @todo Refactor to make it more testable
 * @return bool
 */
	public function rebuildTotals() {
		$this->getBasket(true);

		// process subtotal
		// process Tax
		$subtotal = 0;
		$tax = 0;
		$BasketItems = $this->basketCache['BasketItem'];

		foreach ($BasketItems as $BasketItem) {
			$subtotal += number_format($BasketItem['row_total'], 2, '.', '');
			$tax += number_format($BasketItem['row_tax_amount'], 2, '.', '');
		}

		//Get subtotal + tax now before tax is changed
		$subtotalIncTax = $subtotal + $tax;

		// check for a discount row and amend tax
		$discount = $this->findTotalRow(Configure::read('EvBasket.labels.discount'));

		//Ensure all discount values are positive for discount calculations, returned to negative when saving into the basket
		$discount['amount'] = abs($discount['amount']);
		$discount['tax'] = abs($discount['tax']);
		$discount['display_inc_tax'] = abs($discount['display_inc_tax']);
		$discount['display_ex_tax'] = abs($discount['display_ex_tax']);

		//Update the tax to have the discounted reduction.
		if (
			(float)$subtotal > 0 &&
			! empty($discount['amount']) &&
			(float)$discount['amount'] > 0
		) {
			$this->manageTotalRow(
				Configure::read('EvBasket.labels.discount'),
				-$discount['amount'],
				$discount['sequence'],
				$discount['display_only'],
				-$discount['display_ex_tax'],
				-$discount['display_inc_tax']
			);

			$tax -= $discount['tax'];
		}

		// check for a voucher row and amend tax
		$voucher = $this->findTotalRow(Configure::read('EvBasket.labels.voucher'));
		$voucher['amount'] = -$voucher['amount'];

		//Update the tax to have the voucher reduction.
		if (
			(float)$subtotal > 0 &&
			! empty($voucher['amount']) &&
			(float)$voucher['amount'] > 0
		) {
			$this->manageTotalRow(
				Configure::read('EvBasket.labels.voucher'),
				-$voucher['amount'],
				$voucher['sequence'],
				$voucher['display_only'],
				-$voucher['display_ex_tax'],
				-$voucher['display_inc_tax']
			);

			$tax -= $voucher['tax'];
		}

		// check for delivery placeholder, only add if we haven't added delivery already
		$deliveryPlaceholder = Configure::read('EvBasket.deliveryPlaceholder');
		$delivery = $this->findTotalRow(Configure::read('EvBasket.labels.delivery'));

		if (! isset($deliveryAmount) && isset($delivery['amount'])) {
			$deliveryAmount = $delivery['amount'];
		}

		// check for delivery taxation
		$delTax = 0;
		$deliveryTax = Configure::read('SiteSetting.ev_basket.delivery_tax_rate');
		if ($deliveryTax !== false && $deliveryTax > 0 && isset($deliveryAmount)) {
			$delTax = (($deliveryTax / 100) * $deliveryAmount);
			$tax += $delTax;
		}

		//If a placeholder exists and the delivery isn't set then set the delivery charge to the placeholder
		if ($deliveryPlaceholder !== false && empty($delivery['id'])) {
			// add / update the delivery row
			$this->manageTotalRow(
				Configure::read('EvBasket.labels.delivery'),
				$deliveryPlaceholder,
				20,
				false,
				$deliveryPlaceholder,
				$deliveryPlaceholder + $delTax
			);

			$deliveryAmount = $deliveryPlaceholder;
		}

		// if at this point tax has worked out to be negative - zero it
		if ($tax < 0) {
			$tax = 0;
		}

		// add / update the subtotal row
		$this->manageTotalRow(
			Configure::read('EvBasket.labels.subtotal'),
			$subtotal,
			10,
			true,
			$subtotal,
			$subtotalIncTax
		);

		// add / update the tax row
		$this->manageTotalRow(
			Configure::read('EvBasket.labels.tax'),
			$tax,
			30,
			false
		);

		// call out to recompile all the basket totals
		$totalResult = $this->BasketTotal->recompileTotals($this->basketCache['Basket']['id']);

		if ($totalResult === true) {
			// dispatch totals recompiled
			$this->_controller->getEventManager()->dispatch(
				new CakeEvent('EvBasket.Component.Basket.postTotalRecompile', $this, array(
					'basketId' => $this->basketCache['Basket']['id']
				))
			);

			// rebuild the basket cache
			$this->getBasket(true);

			return true;
		}

		return false;
	}

/**
 * if we're updating an item - check if we actually want to update multi rows in one go
 *
 * @param 	array 		$item 		Array with two elements of model and model_id / or multidimensional array of model, model_id, quantity, unitPrice
 * @param 	int 		$quantity	Amount of items to add
 * @param 	decimal 	$unitPrice 	Unit price of the item
 * @param 	decimal		$taxRate 	The tax rate to update
 * @return  array 		$toUpdate 	Array of items to update
 */
	public function buildUpdateList($item, $quantity, $unitPrice, $taxRate) {
		if (is_numeric(key($item))) {
			return $item;
		}

		$item['quantity'] = $quantity;
		$item['unitPrice'] = $unitPrice;
		$item['taxRate'] = $taxRate;

		return array(
			$item
		);
	}

/**
 * when adding an item, check we have all the data needed
 *
 * @param 	array 	$basketItem 	Array of data - array of model, model_id, quantity, unitPrice
 * @return 	bool
 */
	public function checkAddItem($item) {
		$Model = EvClassRegistry::init($item['BasketItem']['model']);

		if (
			! empty($item['BasketItem']['model']) &&
			! empty($item['BasketItem']['model_id']) &&
			! empty($item['BasketItem']['quantity']) &&
			method_exists($Model, 'getUnitPrice')
		) {
			// check we have enough stock
			if (! $this->hasEnoughStock(EvClassRegistry::init($item['BasketItem']['model']), $item)) {
				$this->_controller->Flash->fail(
					Configure::read('EvInventory.notEnoughStockMessage')
				);
				return false;
			}

			return true;
		}

		return false;
	}

/**
 * check to see if the item has enough stock (assuming inventories are enabled)
 *
 * @param 	object 	$Model 	The model object of the item we are checking
 * @param 	array 	$item 	The basket item
 * @return 	bool
 */
	public function hasEnoughStock($Model, $item) {
		if (CakePlugin::loaded('EvInventory') && $Model->hasBehavior('EvInventory.Inventories')) {
			App::uses('InventoryLib', 'EvInventory.Lib');

			if (! isset($item['BasketItem'])) {
				$item['BasketItem'] = $item;
			}

			if (! InventoryLib::hasEnoughStock($Model->readForEdit($item['BasketItem']['model_id']), $item['BasketItem']['quantity'])) {
				return false;
			}
		}

		return true;
	}

/**
 * find a total row given the name
 *
 * @param 	string 		$totalName 		Row name for the total
 * @return 	array|bool 	$BasketTotal	BasketTotal array row
 */
	public function findTotalRow($name) {
		$this->getBasket();

		if (isset($this->basketCache['BasketTotal'])) {
			$rows = Hash::combine(
				$this->basketCache['BasketTotal'],
				'{n}.name',
				'{n}.id'
			);
		}

		if (isset($rows[$name]) && ! empty($rows[$name])) {
			return $this->basketCache['BasketTotal'][$rows[$name]];
		}

		return false;
	}

/**
 * reset a basket - clear the hash and basket
 *
 * @return 	bool
 */
	public function resetBasket() {
		$this->hash = null;
		$this->basketCache = null;

		return true;
	}

/**
 * get a single basket item
 *
 * @param 	string|int|array 	$key 	Either Basket row ID / compiled basket key string / Array with two elements of model and model_id
 * @return  array|bool 			$item 	Item array or bool false
 */
	public function getBasketItem($key) {
		$this->getBasket();

		// set the key as the itemKey as a fallback
		$itemKey = $this->getItemKey($key);

		if ($this->basketItemExists($itemKey)) {
			return $this->basketCache['BasketItem'][$itemKey];
		}

		return false;
	}

/**
 * get the basket
 *
 * @param 	bool 	$rebuild	Whether to force a rebuild of the basket or not
 * @return 	array 	$basket 	Basket Data
 */
	public function getBasket($rebuild = false) {
		$this->setupHash();
		$this->setupBasket($this->hash);
		$itemContain = $this->buildItemContain(
			Configure::read('EvBasket.BasketItemContains'),
			Configure::read('EvBasket.BasketItemBelongsTo')
		);

		$this->basketCache = $this->Basket->getFullBasket(
			$this->hash,
			$itemContain
		);

		return $this->basketCache;
	}

/**
 * add an item to the basket
 *
 * @param 	array 		$item 		Array with two elements of model and model_id / or multidimensional array of model, model_id, quantity, unitPrice, taxRate
 * @param 	decimal 	$unitPrice 	Unit price of the item
 * @param 	int 		$quantity	Amount of items to add
 * @param 	decimal		$taxRate 	The tax rate to add
 * @param 	bool 		$dispatchEvent 	Whether to disaptch the itemAdd event - used to prevent loops
 * @return 	bool
 */
	public function addItem($items, $unitPrice = 0, $quantity = 1, $taxRate = 0, $dispatchEvent = true) {
		$this->getBasket();

		$toUpdate = $this->buildUpdateList($items, $quantity, $unitPrice, $taxRate);

		$result = false;
		$updated = array();

		foreach ($toUpdate as $item) {

			$itemKey = $this->buildItemKey($item);

			// check if we are updating an existing item or creating a new row
			if ($this->basketItemExists($itemKey)) {
				$itemAdded = $this->BasketItem->updateItem(
					$this->basketCache['BasketItem'][$itemKey],
					($this->basketCache['BasketItem'][$itemKey]['quantity'] + $item['quantity']),
					$item['unitPrice'],
					$item['taxRate'],
					$item['unitTax']
				);
			} else {
				$itemAdded = $this->BasketItem->addItem(
					$this->basketCache['Basket']['id'],
					$item,
					$item['unitPrice'],
					$item['quantity'],
					$item['taxRate']
				);
			}

			if ($itemAdded === true) {
				$result = true;
				$updatedItems[] = $item;
			}
		}

		if ($result === true) {

			//Rebuild totals before event fired to make sure that the totals are up to date before
			// any changes are made.
			$returnTotals = $this->rebuildTotals();
			if ($dispatchEvent) {
				// dispatch item added to basket event
				$this->_controller->getEventManager()->dispatch(
					new CakeEvent('EvBasket.Component.Basket.itemAdd', $this, array(
						'basketId' => $this->basketCache['Basket']['id'],
						'items' => $updatedItems
					))
				);
			}

			return $returnTotals;
		}

		return false;
	}

/**
 * update an item in the basket
 *
 * @param 	array 		$item 		Array with two elements of model and model_id / or multidimensional array of model, model_id, quantity, unitPrice, taxRate
 * @param 	int 		$quantity	Amount of items to add
 * @param 	decimal 	$unitPrice 	Unit price of the item
 * @param 	decimal		$taxRate 	The tax rate to update
 * @param 	bool 		$dispatchEvent 	Whether to disaptch the itemAdd event - used to prevent loops
 * @return 	bool
 */
	public function updateItem($items, $quantity = null, $unitPrice = null, $taxRate = 0, $dispatchEvent = true) {
		$this->getBasket();

		$toUpdate = $this->buildUpdateList($items, $quantity, $unitPrice, $taxRate);

		$result = false;
		$updated = array();

		foreach ($toUpdate as $item) {
			$itemKey = $this->buildItemKey($item);

			// check if the row actually exists
			if (! $this->basketItemExists($itemKey)) {
				continue;
			}

			// check if we're wanting to delete
			if ($item['quantity'] <= 0) {
				$this->deleteItem($item);
				continue;
			}

			// check we have enough stock
			if (! $this->hasEnoughStock(EvClassRegistry::init($item['model']), $item)) {
				$this->_controller->Flash->fail(
					Configure::read('EvInventory.notEnoughStockMessage')
				);
				continue;
			}

			$updateItem = $this->BasketItem->updateItem(
				$this->basketCache['BasketItem'][$itemKey],
				$item['quantity'],
				(isset($item['unitPrice'])) ? $item['unitPrice'] : null,
				(isset($item['taxRate'])) ? $item['taxRate'] : null
			);

			if ($updateItem === true) {
				$result = true;
				$updatedItems[] = $item;
			}
		}

		if ($result === true) {

			//Rebuild totals before event fired to make sure that the totals are up to date before
			// any changes are made.
			$returnTotals = $this->rebuildTotals();
			if ($dispatchEvent) {
				// dispatch item added to basket event
				$this->_controller->getEventManager()->dispatch(
					new CakeEvent('EvBasket.Component.Basket.itemUpdated', $this, array(
						'basketId' => $this->basketCache['Basket']['id'],
						'items' => $updatedItems
					))
				);
			}

			return $returnTotals;
		}

		return false;
	}

/**
 * delete and item from the basket
 *
 * @param 	array 		$item 		Array with two elements of model and model_id
 * @param 	bool 		$dispatchEvent 	Whether to disaptch the itemAdd event - used to prevent loops
 * @return 	bool
 */
	public function deleteItem($item, $dispatchEvent = true) {
		$this->getBasket();

		$itemKey = $this->buildItemKey($item);

		// check if the row actually exists
		if (! $this->basketItemExists($itemKey)) {
			return false;
		}

		$result = $this->BasketItem->deleteItem(
			$this->basketCache['BasketItem'][$itemKey]
		);

		if ($result === true) {

			//Rebuild totals before event fired to make sure that the totals are up to date before
			// any changes are made.
			$returnTotals = $this->rebuildTotals();
			if ($dispatchEvent) {
				// dispatch item added to basket event
				$this->_controller->getEventManager()->dispatch(
					new CakeEvent('EvBasket.Component.Basket.itemDeleted', $this, array(
						'basketId' => $this->basketCache['Basket']['id'],
						'itemKey' => $itemKey,
						'BasketItem' => $this->BasketItem
					))
				);
			}

			return $returnTotals;
		}

		return false;
	}

	public function deleteTotalRow($name) {
		$this->getBasket();

		if (empty($name)) {
			return false;
		}

		$totalRow = $this->findTotalRow($name);

		$basketTotalId = null;
		if ($totalRow !== false && ! empty($totalRow['id'])) {
			$basketTotalId = $totalRow['id'];
		}

		return $this->BasketTotal->deleteTotalRow(
			$basketTotalId
		);
	}

/**
 * manage a total row
 *
 * @param 	string 	$name 			The name label for the row
 * @param 	decimal $amount 		The amount to add
 * @param 	int 	$sequence 		The sequence order number
 * @param 	bool 	$displayOnly 	Whether it's a displayOnly row (not added n grand total calcs)
 * @param 	bool 	$amountExTax 	The amount not including tax to display
 * @param 	bool 	$amountIncTax 	The amount including tax to display
 * @param   decimal	$tax 			Optional field for adding tax to a row total instead of calculating on the fly
 * @return 	bool
 */
	public function manageTotalRow($name, $amount, $sequence = null, $displayOnly = false, $amountExTax = null, $amountIncTax = null) {
		$this->getBasket();

		if (empty($name)) {
			return false;
		}

		$totalRow = $this->findTotalRow($name);

		$data = array(
			'name' => $name,
			'amount' => $amount,
			'display_only' => $displayOnly,
			'display_ex_tax' => $amountExTax,
			'display_inc_tax' => $amountIncTax,
			'tax' => $amountIncTax - $amountExTax
		);

		if (!is_null($amountExTax) && !empty($amountExTax)) {
			$data['display_ex_tax'] = $amountExTax;
		}

		if (!is_null($amountIncTax) && !empty($amountIncTax)) {
			$data['display_in_tax'] = $amountIncTax;
		}

		if (is_null($sequence) && ! empty($totalRow['sequence'])) {
			$data['sequence'] = $totalRow['sequence'];
		} elseif (! empty($sequence)) {
			$data['sequence'] = $sequence;
		}

		$basketTotalId = null;
		if ($totalRow !== false && ! empty($totalRow['id'])) {
			$basketTotalId = $totalRow['id'];
		}

		return $this->BasketTotal->manageTotalRow(
			$this->basketCache['Basket']['id'],
			$data,
			$basketTotalId
		);
	}

/**
 * get basket summary for headers / sidebar
 *
 * @return 	array 	$summary 	summary of amount and no. items
 */
	public function getBasketSummary() {
		$this->getBasket();

		$noItems = 0;
		$total = 0;

		if (isset($this->basketCache['BasketItem']) && ! empty($this->basketCache['BasketItem'])) {
			foreach ($this->basketCache['BasketItem'] as $BasketItem) {
				$noItems += $BasketItem['quantity'];
				$total += $BasketItem['row_total'];
			}
		}

		return array(
			'total' => $total,
			'items' => $noItems
		);
	}

/**
 * build transaction items from basket
 *
 * @return 	array 	$transactionItems
 */
	public function buildTransactionItems() {
		$this->getBasket();

		$items = array();
		foreach ($this->basketCache['BasketItem'] as $BasketItem) {
			$items[] = array(
				'model' => $BasketItem['model'],
				'model_id' => $BasketItem['model_id'],
				'description' => BasketLib::getItemName($BasketItem),
				'amount' => $BasketItem['row_total'],
				'quantity' => $BasketItem['quantity'],
				'unitPrice' => $BasketItem['unit_price']
			);
		}

		return $items;
	}

/**
 * Build the transaction extras array from provided addresses, user and additional data.
 * The totals are found and added to the extras as well.
 * @param  array $addresses The delivery and billing addresses to be used for this transaction
 * @param  array $user      The current user to be used for this transaction
 * @param  array $data      Additional provided data that can be used to add custom extras
 * @return array            The array to be used as transaction extras
 */
	public function buildTransactionExtras($addresses, $user, $data) {
		return [
			'totals' => $this->buildTransactionTotals(),
			'delivery' => (!empty($addresses['delivery'])) ? $addresses['delivery'] : null,
			'user' => $user
		];
	}

/**
 * build transaction totals from basket
 *
 * @return 	array 	$transactionTotals
 */
	public function buildTransactionTotals() {
		$this->getBasket();

		$grandTotal = $this->findTotalRow(Configure::read('EvBasket.labels.grandtotal'));
		$subTotal = $this->findTotalRow(Configure::read('EvBasket.labels.subtotal'));
		$taxTotal = $this->findTotalRow(Configure::read('EvBasket.labels.tax'));
		$deliveryTotal = $this->findTotalRow(Configure::read('EvBasket.labels.delivery'));
		$discount = $this->findTotalRow(Configure::read('EvBasket.labels.discount'));

		return array(
			'delivery' => $deliveryTotal['amount'],
			'tax' => $taxTotal['amount'],
			'subtotal' => $subTotal['amount'],
			'grandtotal' => $grandTotal['amount'],
			'discount' => (! empty($discount['amount'])) ? $discount['amount'] : false
		);
	}

/**
 * set a basket data item
 *
 * @param string   $name          name of the data row
 * @param string   $data          value of the data row
 * @param boolean  $dispatchEvent if true, an event will be dispatched after completing the add/update
 */
	public function setBasketData($name, $data, $isVisible = false, $dispatchEvent = true) {
		$this->getBasket();

		$result = false;

		$dataRecord = $this->BasketData->setData(
			$this->basketCache['Basket']['id'],
			$name,
			$data,
			$isVisible
		);

		if ($dataRecord === true) {
			$result = true;
		}

		if ($result === true) {
			if ($dispatchEvent) {
				// dispatch data record added to basket event
				$this->_controller->getEventManager()->dispatch(
					new CakeEvent('EvBasket.Component.Basket.dataAdd', $this, array(
						'basketId' => $this->basketCache['Basket']['id'],
						'data' => $dataRecord
					))
				);
			}

			// call out to recompile all the basket totals
			return $this->rebuildTotals();
		}

		return false;
	}

/**
 * Get the basket data for the current basket
 * @return array $basketData The data for the current basket
 */
	public function getBasketData() {
		$basketData = array();

		$this->getBasket();

		if (! empty($this->basketCache['BasketData'])) {
			foreach ($this->basketCache['BasketData'] as $dataRow) {
				$basketData[] = array(
					'name' => $dataRow['name'],
					'data' => $dataRow['data'],
					'is_visible' => $dataRow['is_visible']
				);
			}
		}

		return $basketData;
	}

/**
 * Delete an single piece of data from the current basket. If the data is deleted successfully, the
 * dataDeleted event is dispatched.
 * @param  string $name          The name of the data to be deleted e.g. discount_code
 * @param  bool   $dispatchEvent By default deleting data fires an event, set to false to stop event from firing
 * @return bool                  If the data was deleted or not
 */
	public function deleteData($name, $dispatchEvent = true) {
		$this->getBasket();

		// check if the row actually exists
		$basketDataNames = Hash::combine($this->basketCache, 'BasketData.{n}.name', 'BasketData.{n}');

		if (!in_array($name, array_keys($basketDataNames))) {
			return false;
		}

		$dataToDelete = $basketDataNames[$name];

		$result = $this->BasketData->deleteData(
			$dataToDelete
		);

		if ($result === true) {

			//Rebuild totals before event fired to make sure that the totals are up to date before
			// any changes are made.
			$returnTotals = $this->rebuildTotals();
			if ($dispatchEvent) {
				// dispatch item added to basket event
				$this->_controller->getEventManager()->dispatch(
					new CakeEvent('EvBasket.Component.Basket.dataDeleted', $this, array(
						'basketId' => $this->basketCache['Basket']['id'],
						'dataName' => $name,
						'BasketData' => $this->BasketData
					))
				);
			}

			return $returnTotals;
		}

		return false;
	}

/**
 * Check that the basket in it's current state is acceptable to purchase. At it's most basic, just check that there are
 * currently items in the basket so that a user doesn't purchase an empty basket.
 * @return bool Returns true if the basket is purchaseable, otherwise returns false or a message about the reason why
 *                      the basket isn't currently purchasable.
 */
	public function isBasketPurchasable() {
		$basket = $this->getBasket();

		if (count($basket['BasketItem']) <= 0) {
			return 'Your basket is empty. Please check the contents of your basket before making a purchase.';
		}

		return true;
	}

/**
 * Update the currency of the basket. To make sure that the basket is updated correctly, all the items are removed and
 * added again after the currency has been changed. Data is also removed and added again in case the data is dependent
 * on the currency e.g. discounts.
 * @param  int $newCurrencyId The ID of the currency to change the basket too.
 */
	public function updateCurrency($newCurrencyId) {
		$basket = $this->getBasket();
		if ($basket['Basket']['currency_id'] != $newCurrencyId) {
			$basketItems = [];
			foreach ($basket['BasketItem'] as $item) {
				//Get the info needed to add the item back to the basket later
				$saveItem['model'] = $item['model'];
				$saveItem['model_id'] = $item['model_id'];
				$saveItem['quantity'] = $item['quantity'];

				$basketItems[] = $saveItem;

				//Remove the item from the current basket
				//This happens when the currency is changed
			}

			//Get data
			$basketData = [];
			foreach ($basket['BasketData'] as $data) {
				//Get the info needed to add the data back to the basket later
				$saveData['name'] = $data['name'];
				$saveData['data'] = $data['data'];
				$saveData['is_visible'] = $data['is_visible'];

				$basketData[] = $saveData;

				//Remove the data form the current basket
				//This happens when the currency is changed
			}

			//update currency
			$this->_controller->Currencies->setCurrency($newCurrencyId);

			//Add data
			if (!empty($basketData)) {
				foreach ($basketData as $data) {
					//Add the data back to the bask
					if (!$this->setBasketData(
							$data['name'],
							$data['data'],
							$data['is_visible'],
							true
						)
					) {
						//Failed to add data back to basket
						$this->_controller->Flash->fail(
							'Failed to add ' . $data['name'] . 'to the basket.',
							array(
								'key' => 'basket-flash'
							)
						);
					}
				}
			}

			//Add items with quantities
			if (!empty($basketItems)) {
				foreach ($basketItems as $item) {
					//Get the new information for the item
					$Model = EvClassRegistry::init($item['model']);

					$unitPrice = $Model->getUnitPrice($item['model_id']);

					// if there's a method to get the taxRate, do so
					$taxRate = 0;
					if (method_exists($Model, 'getTaxRate')) {
						$taxRate = $Model->getTaxRate($item['model_id']);
					}

					if ($unitPrice !== false) {
						// add the item back to the basket
						if (!$this->addItem($item, $unitPrice, $item['quantity'], $taxRate)) {
							$this->_controller->Flash->fail(
								'Failed to add ' . $item['model'] . ' ' . $item['model_id'] . 'to the basket',
								array(
									'key' => 'basket-flash'
								)
							);
						}
					} else {
						//Failed to find a price for the item, couldn't add it back to basket
						$this->_controller->Flash->fail(
							'Failed to add ' . $item['model'] . ' ' . $item['model_id'] . 'to the basket',
							array(
								'key' => 'basket-flash'
							)
						);
					}
				}

				//Rebuild
				$this->rebuildTotals();
			}
		}
	}
}
