<?php

App::uses('CakeEventListener', 'Event');
App::uses('DiscountLib', 'EvDiscount.Lib');

class EvBasketCheckForDiscountListener implements CakeEventListener {

	public function implementedEvents() {
		return array(
			'EvDiscount.Component.DiscountCodes.codeAdded' => 'addDiscount',
			'EvBasket.Component.Basket.itemAdd' => 'rebuildDiscount',
			'EvBasket.Component.Basket.itemUpdated' => 'rebuildDiscount',
			'EvBasket.Component.Basket.itemDeleted' => 'rebuildDiscount',
			'EvLoyaltyPoints.Component.Points.usePoints' => 'addLoyaltyDiscount',
		);
	}

	/**
	 * queue up the confirmation emails
	 *
	 * @param   CakeEvent
	 */
	public function addDiscount(CakeEvent $Event, $data = null) {
		$dataCode = $Event->data['data'];
		// Event subject will be the EvDiscount.DiscountCodesComponent
		return $this->rebuildDiscount($Event, $dataCode);
	}

	/**
	 * rebuild discount due to basket amends
	 *
	 * @param   CakeEvent
	 */
	public function rebuildDiscount(CakeEvent $Event, $dataCode = null) {
		if (! CakePlugin::loaded('EvDiscount')) {
			return false;
		}

		$EventSubject = $Event->subject();

		$basket = $EventSubject->_controller->BasketManager->getBasket();

		$priceModifier = $this->_setupPriceModifier($basket);

		$this->_populatePriceModifierBasket($priceModifier, $basket);

		$this->_populatePriceModifierDiscounts($priceModifier, $EventSubject, $dataCode);

		// Make the price modifier apply the discounts and get the results back
		$priceModifier->applyDiscounts();
		$discountedBasket = $priceModifier->basket()->toArray();

		//Default discount position after subtotal. Discount will always contain what is to betaken off of the subtotal.
		//Tax amounts always contain the discounted value so display this after discount.
		$discountSequence = 15;

		// Loop through the discount breakdown and add up those for vouchers and those for discounts separately
		// Add a basket data record to handle the discount row
		$voucherPreTaxTotal = $voucherIncTaxTotal = $discountPreTaxTotal = $discountIncTaxTotal = 0;
		foreach ($discountedBasket['discountBreakdown'] as $discountCode => $discountAmount) {
			// Voucher codes get prefixed with V_ at the top of this function when adding them
			if (strpos($discountCode, "V_") === 0) {
				$EventSubject->_controller->BasketManager->setBasketData('voucher_code', str_replace("V_", "", $discountCode), true);
				$voucherPreTaxTotal += $discountAmount['amount'];
				$voucherIncTaxTotal += $discountAmount['amount'] + $discountAmount['tax'];
			} else {
				$EventSubject->_controller->BasketManager->setBasketData('discount_code', $discountCode, true);
				$discountPreTaxTotal += $discountAmount['amount'];
				$discountIncTaxTotal += $discountAmount['amount'] + $discountAmount['tax'];
			}
		}

		// Add Voucher discount total
		if ($voucherPreTaxTotal > 0) {
			//Don't update the tax as that is dealt with when rebuilding totals
			$EventSubject->_controller->BasketManager->manageTotalRow(
				Configure::read('EvBasket.labels.voucher'),
				-$voucherPreTaxTotal,
				$discountSequence,
				false,
				-$voucherPreTaxTotal,
				-$voucherIncTaxTotal
			);
		} else {
			// No discount is applicable to the basket remove the row
			$EventSubject->_controller->BasketManager->deleteTotalRow(Configure::read('EvBasket.labels.voucher'));
		}

		// Add Discount discount total
		if ($discountPreTaxTotal > 0) {
			//Don't update the tax as that is dealt with when rebuilding totals
			$EventSubject->_controller->BasketManager->manageTotalRow(
				Configure::read('EvBasket.labels.discount'),
				-$discountPreTaxTotal,
				$discountSequence,
				false,
				-$discountPreTaxTotal,
				-$discountIncTaxTotal
			);
		} else {
			// No discount is applicable to the basket remove the row
			$EventSubject->_controller->BasketManager->deleteTotalRow(Configure::read('EvBasket.labels.discount'));
		}

		// Now we have added the discount row we can rebuild the totals
		$Event->subject()->_controller->BasketManager->rebuildTotals();

		return true;
	}

	public function addLoyaltyDiscount(CakeEvent $Event) {
		if (! CakePlugin::loaded('EvLoyaltyPoints')) {
			return false;
		}

		$EventSubject = $Event->subject();

		if (! is_object($EventSubject->_controller->BasketManager)) {
			$EventSubject->_controller->BasketManager = $EventSubject->_controller->loadComponent('EvBasket.BasketManager');
		}

		$total = $EventSubject->_controller->BasketManager->findTotalRow(
			Configure::read('EvBasket.labels.subtotal')
		);

		App::uses('PointLib', 'EvLoyaltyPoints.Lib');

		if ($Event->data['pointsToUse'] <= 0) {
			$discount = 0;
		} else {
			$discount = PointLib::calculate($total['amount'], 'EvBasket', 0, $Event->data['pointsToUse']);
		}

		$EventSubject->_controller->BasketManager->manageTotalRow(
			Configure::read('EvBasket.labels.discount'),
			-$discount,
			15
		);

		// Add a basket data record to handle the loyalty disount row
		$EventSubject->_controller->BasketManager->setBasketData('loyalty_points_used', $Event->data['pointsToUse'], true);

		$Event->subject()->_controller->BasketManager->rebuildTotals();

		return true;
	}

/**
 * Sets up the price modifier configuration and returns a setup version
 * @param  array  			$basket Basket array
 * @return PriceModifier    Initialised price modifier object
 */
	protected function _setupPriceModifier($basket) {
		$priceModifier = new Evoluted\PriceModifier\PriceModifier(
			new Evoluted\PriceModifier\Basket\Basket(new Evoluted\PriceModifier\Storage\Runtime, $basket['Basket']['id'])
		);

		require_once (CakePlugin::path('EvDiscount') . 'Lib' . DS . 'DiscountRestrictions.php');
		$priceModifier->setDiscountRestrictionHandler('Evoluted\EvDiscount\DiscountRestrictions\DiscountRestrictions');

		// In the case of our shop we want vouchers to actually go past the subtotal amount so they apply to delivery as well.
		require_once (CakePlugin::path('EvDiscount') . 'Lib' . DS . 'VoucherDiscountModifier.php');
		$priceModifier->addDiscountModifier('voucher', 'Evoluted\PriceModifier\DiscountModifiers\VoucherDiscountModifier');

		// Set whether the basket is an inc or exvat basket
		$priceModifier->applyDiscountsAfterTax = Configure::read('SiteSetting.ev_basket.discounts_include_tax');

		return $priceModifier;
	}

/**
 * Populates the price modifier basket with basket items. Modifies the price modifier directly
 * @param  PriceModifier 	&$priceModifier Price modifier object
 * @param  array 			$basket         Basket Array
 * @return null
 */
	protected function _populatePriceModifierBasket(&$priceModifier, $basket) {
		// Add each basket item to the price modifier
		foreach ($basket['BasketItem'] as $basketItem) {
			$priceModifier->basket()->insert([
				'id' => $basketItem['id'],
				'unitPrice' => $basketItem['unit_price'],
				'quantity' => $basketItem['quantity'],
				'taxRate' => $basketItem['tax_rate'],
				'model' => $basketItem['model'],
				'model_id' => $basketItem['model_id'],
				'Variant' => $basketItem['Variant']
			]);
		}
	}

/**
 * Populates the price modifier basket with basket items. Modifies the price modifier directly
 * @param  PriceModifier 	&$priceModifier Price modifier object
 * @param  array 			$basket         Basket Array
 * @return null
 */
	protected function _populatePriceModifierDiscounts(&$priceModifier, &$EventSubject, $dataCode) {
		// The cookie isn't used any more but there can still be multiple discount codes added to the basket in the
		// database potentially. Current functionality only allows one discount code at a time though.
		if (! is_object($EventSubject->_controller->BasketManager)) {
			$EventSubject->_controller->BasketManager = $EventSubject->_controller->loadComponent('EvBasket.BasketManager');
		}

		// Get the basket data to find the discount codes
		$codes = array();
		if ($dataCode != null) {
			$codes[$dataCode['DiscountCode']['code']] = $dataCode;
		}

		// Add additional discount codes from basket data
		$basketData = $EventSubject->_controller->BasketManager->getBasketData();
		if (!empty($basketData)) {
			foreach ($basketData as $key => $value) {
				if ($value['name'] == 'discount_code' || $value['name'] == 'voucher_code') {
					$codes[$value['data']] = $value;
				}
			}
		}

		// If the basket data returns some codes or one is passed through (or both) then
		if (! empty($codes) && is_array($codes)) {
			// Checks with the discount module to see what codes are active and not run out etc.
			// Only returns codes that are still valid
			$DiscountCode = EvClassRegistry::init('EvDiscount.DiscountCode');
			$validCodes = $DiscountCode->validateCodes($codes);

			if (! empty($validCodes)) {
				// If there were any valid discount modifiers add to the discount rules for the price modifier.
				foreach ($validCodes as $code => $codeData) {
					// Add discount rule
					if ($codeData['DiscountCode']['is_voucher']) {
						$discountType = 'voucher';
					} else {
						$discountType = $codeData['DiscountCode']['discount_type_id'] == 1 ? 'fixed' : 'percentage';
					}

					$priceModifier->addDiscountRule([
						'id' => $codeData['DiscountCode']['is_voucher'] ? ("V_" . $code) : $code,
						'discountType' => $discountType,
						// Used for percentage discounts
						'percent' => $codeData['DiscountCode']['amount'],
						// Used for fixed discounts (ignored by percentage so fine to always pass)
						'amount' => $codeData['DiscountCode']['amount'],
						// Percentage discounts are applied per item and fixed discounts are applied at the end
						'applyToItems' => $codeData['DiscountCode']['discount_type_id'] == 1 ? false : true,
						// Apply discounts before or after applying tax
						'applyDiscountsAfterTax' => Configure::read('SiteSetting.ev_basket.discounts_include_tax'),
						// Additional information passed through for the price modifiers restrictions
						'min_order' => $codeData['DiscountCode']['min_order'],
						'code_restrictions' => $codeData['CodeRestriction']
					]);
				}
			}
		}
	}
}
